"""AnalogVNN: A fully modular framework for modeling and optimizing analog/photonic neural networks."""

import sys

if sys.version_info[:2] >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata  # pragma: no cover

__package__ = 'analogvnn'
__author__ = 'Vivswan Shah (vivswanshah@pitt.edu)'

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    __version__ = '0.0.0'

if sys.version_info < (3, 7, 0):
    import warnings

    warnings.warn(
        'The installed Python version reached its end-of-life. '
        'Please upgrade to a newer Python version for receiving '
        'further gdshelpers updates.', Warning)



import torch
from torch import nn, Tensor

from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.fn.reduce_precision import stochastic_reduce_precision
from analogvnn.nn.precision.Precision import Precision
from analogvnn.utils.common_types import TENSOR_OPERABLE

__all__ = ['StochasticReducePrecision']


class StochasticReducePrecision(Precision, BackwardIdentity):
    """Implements the stochastic reduce precision function.

    Attributes:
        precision (nn.Parameter): the precision of the output tensor.
    """

    __constants__ = ['precision']
    precision: nn.Parameter

    def __init__(self, precision: int = 8):
        """Initialize the StochasticReducePrecision module.

        Args:
            precision (int): the precision of the output tensor.
        """

        super().__init__()
        if precision < 1:
            raise ValueError(f'precision has to be more than 0, but got {precision}')

        if precision != int(precision):
            raise ValueError(f'precision must be int, but got {precision}')

        self.precision = nn.Parameter(torch.tensor(precision), requires_grad=False)

    @property
    def precision_width(self) -> Tensor:
        """The precision width.

        Returns:
            Tensor: the precision width
        """

        return 1 / self.precision

    @property
    def bit_precision(self) -> Tensor:
        """The bit precision of the ReducePrecision module.

        Returns:
            Tensor: the bit precision of the ReducePrecision module.
        """

        return torch.log2(self.precision + 1)

    @staticmethod
    def convert_to_precision(bit_precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """Convert the bit precision to the precision.

        Args:
            bit_precision (TENSOR_OPERABLE): the bit precision.

        Returns:
            TENSOR_OPERABLE: the precision.
        """

        return 2 ** bit_precision - 1

    def extra_repr(self) -> str:
        """The extra __repr__ string of the StochasticReducePrecision module.

        Returns:
            str: string
        """

        return f'precision={int(self.precision)}'

    def forward(self, x: Tensor) -> Tensor:
        """Forward function of the StochasticReducePrecision module.

        Args:
            x (Tensor): input tensor.

        Returns:
            Tensor: output tensor.
        """

        return stochastic_reduce_precision(x, self.precision)



from __future__ import annotations

import functools
import inspect
from typing import Union, Type, Callable, Sequence, Optional, Set, Iterator, Tuple

from torch import nn, Tensor

from analogvnn.backward.BackwardFunction import BackwardFunction
from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.graph.ArgsKwargs import ArgsKwargs, ArgsKwargsOutput
from analogvnn.utils.common_types import TENSORS

__all__ = ['Layer']


# https://github.com/pytorch/pytorch/pull/91819
def __nn_Module_init_updated__(function: Callable) -> Callable:
    """Wrapper for nn.Module.__init__ to support multiple parent classes at same time.

    Args:
        function (Callable): nn.Module.__init__ function

    Returns:
        Callable: Wrapped function
    """

    # noinspection PyUnusedLocal
    def _temp(*args, **kwargs) -> ...:
        pass

    @functools.wraps(function)
    def new_function(self, *args, **kwargs):
        super_init = None
        next_mro_index = self.__class__.__mro__.index(nn.Module) + 1
        next_mro_class = self.__class__.__mro__[next_mro_index]

        if next_mro_class is not object:
            super_init = next_mro_class.__init__
            next_mro_class.__init__ = _temp

        function(self, *args, **kwargs)

        if next_mro_class is not object:
            next_mro_class.__init__ = super_init
            super(nn.Module, self).__init__()

    return new_function


if not hasattr(nn.Module, 'call_super_init'):
    nn.Module.__init__ = __nn_Module_init_updated__(nn.Module.__init__)
    """nn.Module.__init__ is updated to support multiple parent classes at same time. """


class Layer(nn.Module):
    """Base class for analog neural network modules.

    Attributes:
        _inputs (Union[None, ArgsKwargs]): Inputs of the layer.
        _outputs (Union[None, Tensor, Sequence[Tensor]]): Outputs of the layer.
        _backward_module (Optional[BackwardModule]): Backward module of the layer.
        _use_autograd_graph (bool): If True, the autograd graph is used to calculate the gradients.
        call_super_init (bool): If True, the super class __init__ of nn.Module is called
        https://github.com/pytorch/pytorch/pull/91819
    """

    _inputs: Union[None, ArgsKwargs]
    _outputs: Union[None, Tensor, Sequence[Tensor]]
    _backward_module: Optional[BackwardModule]
    _use_autograd_graph: bool

    # https://github.com/pytorch/pytorch/pull/91819
    call_super_init: bool = True

    def __init__(self):
        """Initializes the layer."""
        super().__init__()
        self._inputs = None
        self._outputs = None
        self._backward_module = None
        self._use_autograd_graph = False

    def __call__(self, *inputs, **kwargs):
        """Calls the forward pass of neural network layer.

        Args:
            *inputs: Inputs of the forward pass.
            **kwargs: Keyword arguments of the forward pass.
        """

        self._forward_wrapper(self.forward)
        outputs = super().__call__(*inputs, **kwargs)
        if self.training:
            self._inputs = ArgsKwargs(args=inputs, kwargs=kwargs)
            self._outputs = outputs

        return outputs

    @property
    def use_autograd_graph(self) -> bool:
        """If True, the autograd graph is used to calculate the gradients.

        Returns:
            bool: use_autograd_graph.
        """

        return self._use_autograd_graph

    @use_autograd_graph.setter
    def use_autograd_graph(self, use_autograd_graph: bool):
        """Sets the use_autograd_graph attribute.

        Args:
            use_autograd_graph (bool): use_autograd_graph.
        """

        self._use_autograd_graph = use_autograd_graph

    @property
    def inputs(self) -> ArgsKwargsOutput:
        """Inputs of the layer.

        Returns:
            ArgsKwargsOutput: inputs.
        """

        return ArgsKwargs.from_args_kwargs_object(self._inputs)

    @property
    def outputs(self) -> Union[None, Tensor, Sequence[Tensor]]:
        """Outputs of the layer.

        Returns:
            Union[None, Tensor, Sequence[Tensor]]: outputs.
        """

        return self._outputs

    @property
    def backward_function(self) -> Union[None, Callable, BackwardModule]:
        """Backward module of the layer.

        Returns:
            Union[None, Callable, BackwardModule]: backward_function.
        """

        if self._backward_module is not None:
            return self._backward_module

        if isinstance(self, BackwardModule):
            return self

        return None

    @backward_function.setter
    def backward_function(self, function: Union[BackwardModule, Type[BackwardModule], Callable]):
        """Sets the backward_function attribute.

        Args:
            function (Union[BackwardModule, Type[BackwardModule], Callable]): backward_function.
        """

        self.set_backward_function(function)

    def set_backward_function(self, backward_class: Union[Callable, BackwardModule, Type[BackwardModule]]) -> Layer:
        """Sets the backward_function attribute.

        Args:
            backward_class (Union[Callable, BackwardModule, Type[BackwardModule]]): backward_function.

        Returns:
            Layer: self.

        Raises:
            TypeError: If backward_class is not a callable or BackwardModule.
        """

        if backward_class == self:
            return self

        if inspect.isclass(backward_class) and issubclass(backward_class, BackwardModule):
            self._backward_module = backward_class(self)
        elif isinstance(backward_class, BackwardModule):
            backward_class.set_layer(self)
            self._backward_module = backward_class
        elif callable(backward_class):
            self._backward_module = BackwardFunction(backward_class, self)
        else:
            raise TypeError(f'Backward Module is not set for "{self}"')

        return self

    def named_registered_children(
            self,
            memo: Optional[Set[nn.Module]] = None
    ) -> Iterator[Tuple[str, nn.Module]]:
        """Returns an iterator over immediate registered children modules.

        Args:
            memo: a memo to store the set of modules already added to the result


        Yields:
            (str, Module): Tuple containing a name and child module

        Note:
            Duplicate modules are returned only once. In the following
            example, ``l`` will be returned only once.
        """

        if memo is None:
            memo = set()

        memo.add(self)
        memo.add(self.backward_function)

        for name, module in self.named_children():
            if module in memo:
                continue

            yield name, module

    def registered_children(self) -> Iterator[nn.Module]:
        r"""Returns an iterator over immediate registered children modules.

        Yields:
            nn.Module: a module in the network

        Note:
            Duplicate modules are returned only once. In the following
            example, ``l`` will be returned only once.
        """

        for _, module in self.named_registered_children():
            yield module

    def _forward_wrapper(self, function: Callable) -> Callable:
        """Wrapper for the forward function.

        Args:
            function (Callable): Forward function.

        Returns:
            Callable: Wrapped function.
        """

        # noinspection PyUnresolvedReferences
        if hasattr(function, '__wrapper__') and function.__wrapper__ == Layer._forward_wrapper:
            return function

        if not isinstance(self.backward_function, BackwardModule):
            return function

        if not self.backward_function.has_forward():
            self.backward_function.forward = self.forward

        @functools.wraps(function)
        def new_forward(*args, **kwargs):
            return self.backward_function.auto_apply(
                *args,
                to_apply=self.use_autograd_graph,
                **kwargs
            )

        new_forward.__wrapped__ = function
        new_forward.__wrapper__ = Layer._forward_wrapper
        self.forward = new_forward
        return new_forward

    def _call_impl_forward(self, *args: Tensor, **kwargs: Tensor) -> TENSORS:
        """Calls the forward pass of the layer.

        Args:
            *args: Inputs of the forward pass.
            **kwargs: Keyword arguments of the forward pass.

        Returns:
            TENSORS: Outputs of the forward pass.
        """

        if isinstance(self.backward_function, BackwardModule) and self.backward_function.has_forward():
            forward_functions = self.backward_function.forward
        else:
            forward_functions = self.forward

        if hasattr(forward_functions, '__wrapped__'):
            forward_functions = forward_functions.__wrapped__

        return forward_functions(*args, **kwargs)



# The original snippet is licensed under the MIT License.
# The following code is modified snippet from https://github.com/szagoruyko/pytorchviz/blob/master/torchviz/dot.py
# to render the autograd graph of a module or a tensor for analogvnn.

from __future__ import annotations

import dataclasses
import typing
import warnings
from collections import namedtuple
from dataclasses import dataclass
from distutils.version import LooseVersion
from functools import partial
from pathlib import Path
from typing import Optional, Sequence, List, Dict, Union, Any, Callable, Iterator, Tuple

import torch
from torch import Tensor, nn
from torch.nn import Parameter

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.is_cpu_cuda import is_cpu_cuda

if typing.TYPE_CHECKING:
    from graphviz import Digraph

__all__ = [
    'size_to_str',
    'AutoGradDot',
    'make_autograd_obj_from_outputs',
    'make_autograd_obj_from_module',
    'get_autograd_dot_from_outputs',
    'get_autograd_dot_from_module',
    'get_autograd_dot_from_trace',
    'save_autograd_graph_from_outputs',
    'save_autograd_graph_from_module',
    'save_autograd_graph_from_trace',
]

Node = namedtuple('Node', ('name', 'inputs', 'attr', 'op'))

# Saved attrs for grad_fn (incl. saved variables) begin with `._saved_*`
SAVED_PREFIX = '_saved_'


def size_to_str(size):
    """Convert a tensor size to a string.

    Args:
        size (torch.Size): the size to convert.

    Returns:
        str: the string representation of the size.
    """

    return '(' + ', '.join(['%d' % s for s in size]) + ')'


def _format_name_size(name, size):
    return '%s\n %s' % (name, size)


def resize_graph(dot: Digraph, size_per_element: float = 0.15, min_size: float = 12):
    """Resize the graph according to how much content it contains.

    Modify the graph in place.

    Args:
        dot (graphviz.Digraph): graph to be resized
        size_per_element (float): A "rank" in graphviz contains roughly
            size_per_element**2 pixels.
        min_size (float): Minimum size of graph.
    """

    # Get the approximate number of nodes and edges
    num_rows = len(dot.body)
    content_size = num_rows * size_per_element
    size = max(min_size, content_size)
    size_str = str(size) + ',' + str(size)
    dot.graph_attr.update(size=size_str)


def get_fn_name(fn: Callable, show_attrs: bool, max_attr_chars: int) -> str:
    """Get the name of a function.

    Args:
        fn (Callable): the function.
        show_attrs (bool): whether to show the attributes.
        max_attr_chars (int): the maximum number of characters to show for the attributes.

    Returns:
        str: the name of the function.
    """

    name = str(type(fn).__name__)
    if name.endswith('Backward'):
        name = name[:-8]
    if not show_attrs:
        return name
    attrs = {}
    for attr in dir(fn):
        if not attr.startswith(SAVED_PREFIX):
            continue
        val = getattr(fn, attr)
        attr = attr[len(SAVED_PREFIX):]
        if torch.is_tensor(val):
            attrs[attr] = '[saved tensor]'
        elif isinstance(val, (tuple, list)) and any(torch.is_tensor(t) for t in val):
            attrs[attr] = '[saved tensors]'
        else:
            attrs[attr] = str(val)
    if not attrs:
        return name
    max_attr_chars = max(max_attr_chars, 3)
    col1width = max(len(k) for k in attrs.keys())
    col2width = min(max(len(str(v)) for v in attrs.values()), max_attr_chars)
    sep = '-' * max(col1width + col2width + 2, len(name))
    attrstr = '%-' + str(col1width) + 's: %' + str(col2width) + 's'
    truncate = lambda s: s[:col2width - 3] + '...' if len(s) > col2width else s  # noqa: E731
    params = '\n'.join(attrstr % (k, truncate(str(v))) for (k, v) in attrs.items())
    return name + '\n' + sep + '\n' + params


@dataclass
class AutoGradDot:
    """Stores and manages Graphviz representation of PyTorch autograd graph.

    Attributes:
        dot (graphviz.Digraph): Graphviz representation of the autograd graph.
        _module (nn.Module): The module to be traced.
        _inputs (List[Tensor]): The inputs to the module.
        _inputs_kwargs (Dict[str, Tensor]): The keyword arguments to the module.
        _outputs (Sequence[Tensor]): The outputs of the module.
        param_map (Dict[int, str]): A map from parameter values to their names.
        _seen (set): A set of nodes that have already been added to the graph.
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
        _called (bool): the module has been called.
    """

    dot: Digraph = None
    _module: nn.Module = None

    _inputs: Sequence[Tensor] = dataclasses.field(default=None, repr=False, hash=False)
    _inputs_kwargs: Dict[str, Tensor] = dataclasses.field(default_factory=dict, repr=False, hash=False)
    _outputs: Sequence[Tensor] = dataclasses.field(default=None, repr=False, hash=False)

    param_map: dict = dataclasses.field(default_factory=dict, repr=False, hash=False)
    _seen: set = dataclasses.field(default_factory=set, repr=False, hash=False)

    show_attrs: bool = dataclasses.field(default=False, repr=False, hash=False)
    show_saved: bool = dataclasses.field(default=False, repr=False, hash=False)
    max_attr_chars: int = dataclasses.field(default=50, repr=False, hash=False)
    _called: bool = False
    _ignore_tensor: Dict[int, bool] = dataclasses.field(default_factory=dict, repr=False, hash=False)

    def __post_init__(self):
        """Create the graphviz graph.

        Raises:
            ImportError: if graphviz (https://pygraphviz.github.io/) is not available.
        """

        try:
            from graphviz import Digraph
        except ImportError as e:
            raise ImportError('requires graphviz: https://pygraphviz.github.io/') from e

        node_attr = {
            'style': 'filled',
            'shape': 'box',
            'align': 'left',
            'fontsize': '12',
            'ranksep': '0.1',
            'height': '0.2',
            'fontname': 'monospace'
        }
        self.dot = Digraph(node_attr=node_attr, graph_attr={'size': '12,12'}, format='svg')
        # noinspection PyProtectedMember
        self.add_ignore_tensor(BackwardModule._empty_holder_tensor)

    @property
    def inputs(self) -> Sequence[Tensor]:
        """The arg inputs to the module.

        Returns:
            Sequence[Tensor]: the arg inputs to the module.
        """

        return self._inputs

    @inputs.setter
    def inputs(self, inputs: Union[Tensor, Sequence[Tensor]]):
        """Set the inputs to the module.

        Args:
            inputs (Union[Tensor, Sequence[Tensor]]): the inputs to the module.
        """

        self._inputs = inputs
        self._called = False

        if not inputs:
            return

        if not isinstance(inputs, (tuple, list)):
            inputs = (inputs,)

        for i, v in enumerate(inputs):
            self.param_map[id(v)] = f'INPUT_{i}'
            self.param_map[id(v.data)] = f'INPUT_{i}'

    @property
    def inputs_kwargs(self) -> Dict[str, Tensor]:
        """The keyword inputs to the module.

        Args:
            Dict[str, Tensor]: the keyword inputs to the module.
        """

        return self._inputs_kwargs

    @inputs_kwargs.setter
    def inputs_kwargs(self, inputs_kwargs: Dict[str, Tensor]):
        """Set the keyword inputs to the module.

        Args:
            inputs_kwargs (Dict[str, Tensor]): the keyword inputs to the module.
        """

        self._inputs_kwargs = inputs_kwargs
        self._called = False

        if not inputs_kwargs:
            return

        for k, v in inputs_kwargs.items():
            self.param_map[id(v)] = f'INPUT_{k}'
            self.param_map[id(v.data)] = f'INPUT_{k}'

    @property
    def outputs(self) -> Optional[Sequence[Tensor]]:
        """The outputs of the module.

        Returns:
            Optional[Sequence[Tensor]]: the outputs of the module.
        """

        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._called = True
        if outputs is not None and not isinstance(outputs, (tuple, list)):
            outputs = (outputs,)

        self._outputs = outputs
        for i, v in enumerate(outputs):
            self.param_map[id(v)] = f'OUTPUT_{i}'
            self.param_map[id(v.data)] = f'OUTPUT_{i}'

    @property
    def module(self) -> nn.Module:
        """The module.

        Returns:
            nn.Module: the module to be traced.
        """

        return self._module

    @module.setter
    def module(self, module: nn.Module):
        """Set the module.

        Args:
            module (nn.Module): the module.
        """

        self._module = module
        self._called = False

        if not module:
            return

        assert isinstance(module, nn.Module)
        for k, v in dict(module.named_parameters()).items():
            self.param_map[id(v)] = k
            self.param_map[id(v.data)] = k

    def reset_params(self):
        """Reset the param_map and _seen.

        Returns:
            AutoGradDot: self.
        """

        self.param_map = {}
        self._seen = set()
        self.inputs = self.inputs
        self.inputs_kwargs = self.inputs_kwargs
        self.module = self.module
        return self

    @property
    def ignore_tensor(self) -> Dict[int, bool]:
        """The tensor ignored from the dot graphs.

        Returns:
            Dict[int, bool]: the ignore tensor dict.
        """

        return self._ignore_tensor

    def add_ignore_tensor(self, tensor: Tensor):
        """Add a tensor to the ignore tensor dict.

        Args:
            tensor (Tensor): the tensor to ignore.

        Returns:
            AutoGradDot: self.
        """

        self._ignore_tensor[id(tensor)] = True
        return self

    def del_ignore_tensor(self, tensor: Tensor):
        """Delete a tensor from the ignore tensor dict.

        Args:
            tensor (Tensor): the tensor to delete.

        Returns:
            AutoGradDot: self.
        """

        self._ignore_tensor.pop(id(tensor), None)
        return self

    def get_tensor_name(self, tensor: Tensor, name: Optional[str] = None) -> Tuple[str, str]:
        """Get the name of the tensor.

        Args:
            tensor (Tensor): the tensor.
            name (Optional[str]): the name of the tensor. Defaults to None.

        Returns:
            Tuple[str, str]: the name and size of the tensor.
        """

        if not name:
            if id(tensor) in self.param_map:
                name = self.param_map[id(tensor)]
            elif hasattr(tensor, 'name') and not not tensor.name:
                name = tensor.name
            elif hasattr(tensor, 'names') and not not tensor.names:
                if len(tensor.names) == 1:
                    name = tensor.names[0]
                else:
                    name = str(tensor.names)
            else:
                name = ''
        name, size = name.strip(), size_to_str(tensor.size()).strip()
        return name, size

    def add_tensor(self, tensor: Tensor, name: Optional[str] = None, _attributes=None, **kwargs):
        """Add a tensor to the graph.

        Args:
            tensor (Tensor): the tensor.
            name (Optional[str]): the name of the tensor. Defaults to None.
            _attributes (Optional[Dict[str, str]]): the attributes of the tensor. Defaults to None.
            **kwargs: the attributes of the dot.node function.

        Returns:
            AutoGradDot: self.
        """

        self._seen.add(tensor)
        self.dot.node(
            name=str(id(tensor)),
            label=_format_name_size(*self.get_tensor_name(tensor, name=name)),
            _attributes=_attributes,
            **kwargs
        )
        return self

    def add_fn(self, fn: Any, _attributes=None, **kwargs):
        """Add a function to the graph.

        Args:
            fn (Any): the function.
            _attributes (Optional[Dict[str, str]]): the attributes of the function. Defaults to None.
            **kwargs: the attributes of the dot.node function.

        Returns:
            AutoGradDot: self.
        """

        self._seen.add(fn)
        self.dot.node(
            name=str(id(fn)),
            label=get_fn_name(fn, self.show_attrs, self.max_attr_chars),
            _attributes=_attributes,
            **kwargs
        )
        return self

    def add_edge(self, u: Any, v: Any, label: Optional[str] = None, _attributes=None, **kwargs):
        """Add an edge to the graph.

        Args:
            u (Any): tail node.
            v (Any): head node.
            label (Optional[str]): the label of the edge. Defaults to None.
            _attributes (Optional[Dict[str, str]]): the attributes of the edge. Defaults to None.
            **kwargs: the attributes of the dot.edge function.

        Returns:
            AutoGradDot: self.
        """

        self.dot.edge(
            tail_name=str(id(u)),
            head_name=str(id(v)),
            label=label,
            _attributes=_attributes,
            **kwargs
        )
        return self

    def add_seen(self, item: Any):
        """Add an item to the seen set.

        Args:
            item (Any): the item.

        Returns:
            AutoGradDot: self.
        """

        self._seen.add(item)
        return self

    def is_seen(self, item: Any) -> bool:
        """Check if the item is in the seen set.

        Args:
            item (Any): the item.

        Returns:
            bool: True if the item is in the seen set.
        """

        return item in self._seen


def _add_grad_fn(link: Union[Tensor, Callable], autograd_dot: AutoGradDot) -> Optional[List]:  # noqa: C901
    """Add a link to the graph.

    Args:
        link (Union[Tensor, Callable]): the Tensor or Tensor.grad_fn.
        autograd_dot (AutoGradDot): the AutoGradDot object.
    """

    if autograd_dot.is_seen(link):
        return None

    next_links = []

    if isinstance(link, Tensor):
        tensor = link

        autograd_dot.add_tensor(tensor, fillcolor='darkolivegreen1' if not tensor._is_view() else 'darkolivegreen3')

        if tensor.grad_fn:
            next_links.append(tensor.grad_fn)
            autograd_dot.add_edge(tensor.grad_fn, tensor)

        if tensor._is_view():
            next_links.append(tensor._base)
            autograd_dot.add_edge(tensor._base, tensor, style='dotted')

        return next_links

    grad_fn = link
    # add the node for this grad_fn
    autograd_dot.add_fn(grad_fn)

    if autograd_dot.show_saved:
        for attr in dir(grad_fn):
            if not attr.startswith(SAVED_PREFIX):
                continue

            val = getattr(grad_fn, attr)
            autograd_dot.add_seen(val)
            attr = attr[len(SAVED_PREFIX):]

            if torch.is_tensor(val):
                autograd_dot.add_edge(grad_fn, val, dir='none')
                autograd_dot.add_tensor(val, name=attr, fillcolor='orange')
                continue

            if isinstance(val, (tuple, list)):
                for i, t in enumerate(val):
                    if not torch.is_tensor(t):
                        continue
                    name = attr + '[%s]' % str(i)
                    autograd_dot.add_edge(grad_fn, t, dir='none')
                    autograd_dot.add_tensor(t, name=name, fillcolor='orange')

    if hasattr(grad_fn, 'variable'):
        # if grad_accumulator, add the node for `.variable`
        var = grad_fn.variable
        autograd_dot.add_tensor(var, fillcolor='lightblue')
        autograd_dot.add_edge(var, grad_fn)

    # recurse
    if hasattr(grad_fn, 'next_functions'):
        for u in grad_fn.next_functions:
            if u[0] is None:
                continue

            if (
                    u[0].__class__.__name__ == 'AccumulateGrad' and
                    hasattr(u[0], 'variable') and
                    id(u[0].variable) in autograd_dot.ignore_tensor
            ):
                continue

            autograd_dot.add_edge(u[0], grad_fn)
            next_links.append(u[0])

    # note: this used to show .saved_tensors in pytorch0.2, but stopped
    # working* as it was moved to ATen and Variable-Tensor merged
    # also note that this still works for custom autograd functions
    if hasattr(grad_fn, 'saved_tensors'):
        for t in grad_fn.saved_tensors:
            if t is None:
                continue
            autograd_dot.add_edge(t, grad_fn)
            autograd_dot.add_tensor(t, fillcolor='orange')

    return next_links


def _compile_autograd_obj(
        autograd_dot: AutoGradDot,
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
) -> AutoGradDot:
    """Make dot graph in AutoGradDot.

    If a node represents a backward function, it is gray. Otherwise, the node
    represents a tensor and is either blue, orange, or green:
     - Blue: reachable leaf tensors that requires grad (tensors whose `.grad`
         fields will be populated during `.backward()`)
     - Orange: saved tensors of custom autograd functions as well as those
         saved by built-in backward nodes
     - Green: tensor passed in as outputs
     - Dark green: if any output is a view, we represent its base tensor with
         a dark green node.

    Args:
        autograd_dot (AutoGradDot): the AutoGradDot object.
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.

    Returns:
        AutoGradDot: graphviz representation of autograd graph
    """

    if LooseVersion(torch.__version__) < LooseVersion('1.9') and (show_attrs or show_saved):
        warnings.warn(
            'make_dot: showing grad_fn attributes and saved variables'
            ' requires PyTorch version >= 1.9. (This does NOT apply to'
            ' saved tensors saved by custom autograd functions.)'
        )

    autograd_dot.show_attrs = show_attrs
    autograd_dot.show_saved = show_saved
    autograd_dot.max_attr_chars = max_attr_chars

    if additional_params is not None:
        autograd_dot.param_map.update(additional_params)

    deque = list(autograd_dot.outputs)

    while len(deque) > 0:
        r = _add_grad_fn(deque.pop(0), autograd_dot=autograd_dot)
        if r is not None:
            deque += r

    resize_graph(autograd_dot.dot)

    return autograd_dot


def _toggle_autograd_backward(disable, status, self):
    if not isinstance(self, Layer):
        return

    self = self.backward_function

    if self is None:
        return

    if disable:
        status[id(self)] = self._disable_autograd_backward
        self._disable_autograd_backward = True
    else:
        self._disable_autograd_backward = status[id(self)]


def make_autograd_obj_from_outputs(
        outputs: Union[Tensor, Sequence[Tensor]],
        named_params: Union[Dict[str, Any], Iterator[Tuple[str, Parameter]]],
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
) -> AutoGradDot:
    """Compile Graphviz representation of PyTorch autograd graph from output tensors.

    Args:
        outputs (Union[Tensor, Sequence[Tensor]]): output tensor(s) of forward pass
        named_params (Union[Dict[str, Any], Iterator[Tuple[str, Parameter]]]): dict of params to label nodes with
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.

    Returns:
        AutoGradDot: graphviz representation of autograd graph
    """

    autograd_dot = AutoGradDot()
    autograd_dot.outputs = outputs
    named_params = dict(named_params)

    for k, v in named_params.items():
        autograd_dot.param_map[id(v)] = k
        autograd_dot.param_map[id(v.data)] = k

    return _compile_autograd_obj(autograd_dot, additional_params, show_attrs, show_saved, max_attr_chars)


def make_autograd_obj_from_module(
        module: nn.Module,
        *args: Tensor,
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
        from_forward: bool = False,
        **kwargs: Tensor
) -> AutoGradDot:
    """Compile Graphviz representation of PyTorch autograd graph from forward pass.

    Args:
        module (nn.Module): PyTorch model
        *args (Tensor): input to the model
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
        from_forward (bool): if True then use autograd graph otherwise analogvvn graph
        **kwargs (Tensor): input to the model

    Returns:
        AutoGradDot: graphviz representation of autograd graph
    """

    assert isinstance(module, nn.Module)
    new_args = []
    new_kwargs = {}
    device = is_cpu_cuda.get_module_device(module)

    for i in args:
        assert isinstance(i, Tensor)
        i = i.to(device)
        i = i.detach()
        i = i.requires_grad_(True)
        new_args.append(i)

    for k, v in kwargs.items():
        assert isinstance(v, Tensor)
        v = v.to(device)
        v = v.detach()
        v = v.requires_grad_(True)
        new_kwargs[k] = v

    autograd_dot = AutoGradDot()
    autograd_dot.module = module
    autograd_dot.inputs = new_args
    autograd_dot.inputs_kwargs = new_kwargs

    training_status = module.training
    use_autograd_graph_status = False
    if isinstance(module, Layer):
        use_autograd_graph_status = module.use_autograd_graph
        module.use_autograd_graph = True

    disable_autograd_backward_status = {}
    if from_forward:
        module.apply(partial(_toggle_autograd_backward, True, disable_autograd_backward_status))

    module.train()
    autograd_dot.outputs = module(*new_args, **new_kwargs)
    module.train(training_status)

    if from_forward:
        module.apply(partial(_toggle_autograd_backward, False, disable_autograd_backward_status))

    if isinstance(module, Layer):
        module.use_autograd_graph = use_autograd_graph_status

    autograd_dot = _compile_autograd_obj(autograd_dot, additional_params, show_attrs, show_saved, max_attr_chars)

    return autograd_dot


def _join_scope_name(name: str, scope: Dict[str, str]) -> str:
    return '/'.join([scope[name], name])


def parse_trace_graph(graph) -> List[Node]:
    scope = {}
    for n in graph.nodes():
        inputs = [i.uniqueName() for i in n.inputs()]
        for i in range(1, len(inputs)):
            scope[inputs[i]] = n.scopeName()

        uname = next(n.outputs()).uniqueName()
        assert n.scopeName() != '', f'{n} has empty scope name'
        scope[uname] = n.scopeName()

    scope['0'] = 'input'

    nodes = []
    for n in graph.nodes():
        attrs = {k: n[k] for k in n.attributeNames()}
        attrs = str(attrs).replace("'", ' ')
        inputs = [_join_scope_name(i.uniqueName(), scope) for i in n.inputs()]
        uname = next(n.outputs()).uniqueName()

        nodes.append(Node(
            name=_join_scope_name(uname, scope),
            op=n.kind(),
            inputs=inputs,
            attr=attrs
        ))

    for n in graph.inputs():
        uname = n.uniqueName()
        if uname not in scope.keys():
            scope[uname] = 'unused'

        nodes.append(Node(
            name=_join_scope_name(uname, scope),
            op='Parameter',
            inputs=[],
            attr=str(n.type())
        ))

    return nodes


def get_autograd_dot_from_trace(trace) -> Digraph:
    """Produces graphs of torch.jit.trace outputs.

    Example:
    >>> trace, = torch.jit.trace(model, args=(x,))
    >>> dot = get_autograd_dot_from_trace(trace)

    Args:
        trace (torch.jit.trace): the trace object to visualize.

    Returns:
        graphviz.Digraph: the resulting graph.
    """

    try:
        from graphviz import Digraph
    except ImportError as e:
        raise ImportError('requires graphviz: https://pygraphviz.github.io/') from e

    # from tensorboardX
    if LooseVersion(torch.__version__) >= LooseVersion('0.4.1'):
        torch.onnx._optimize_trace(trace, torch._C._onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    elif LooseVersion(torch.__version__) >= LooseVersion('0.4'):
        torch.onnx._optimize_trace(trace, False)
    else:
        torch.onnx._optimize_trace(trace)
    graph = trace.graph()
    list_of_nodes = parse_trace_graph(graph)

    node_attr = {
        'style': 'filled',
        'shape': 'box',
        'align': 'left',
        'fontsize': '12',
        'ranksep': '0.1',
        'height': '0.2'
    }

    dot = Digraph(node_attr=node_attr, graph_attr={'size': '12,12'})

    for node in list_of_nodes:
        dot.node(node.name, label=node.name.replace('/', '\n'))
        if node.inputs:
            for inp in node.inputs:
                dot.edge(inp, node.name)

    resize_graph(dot)

    return dot


def get_autograd_dot_from_outputs(
        outputs: Union[Tensor, Sequence[Tensor]],
        named_params: Union[Dict[str, Any], Iterator[Tuple[str, Parameter]]],
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
) -> Digraph:
    """Runs and make Graphviz representation of PyTorch autograd graph from output tensors.

    Args:
        outputs (Union[Tensor, Sequence[Tensor]]): output tensor(s) of forward pass
        named_params (Union[Dict[str, Any], Iterator[Tuple[str, Parameter]]]): dict of params to label nodes with
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.

    Returns:
        Digraph: graphviz representation of autograd graph
    """

    return make_autograd_obj_from_outputs(
        outputs=outputs,
        named_params=named_params,
        additional_params=additional_params,
        show_attrs=show_attrs,
        show_saved=show_saved,
        max_attr_chars=max_attr_chars,
    ).dot


def get_autograd_dot_from_module(
        module: nn.Module,
        *args: Tensor,
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
        from_forward: bool = False,
        **kwargs: Tensor
) -> Digraph:
    """Runs and make Graphviz representation of PyTorch autograd graph from forward pass.

    Args:
        module (nn.Module): PyTorch model
        *args (Tensor): input to the model
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
        from_forward (bool): if True then use autograd graph otherwise analogvvn graph
        **kwargs (Tensor): input to the model

    Returns:
        Digraph: graphviz representation of autograd graph
    """

    return make_autograd_obj_from_module(
        module,
        *args,
        additional_params=additional_params,
        show_attrs=show_attrs,
        show_saved=show_saved,
        max_attr_chars=max_attr_chars,
        from_forward=from_forward,
        **kwargs
    ).dot


def save_autograd_graph_from_outputs(
        filename: Union[str, Path],
        outputs: Union[Tensor, Sequence[Tensor]],
        named_params: Union[Dict[str, Any], Iterator[Tuple[str, Parameter]]],
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
) -> str:
    """Save Graphviz representation of PyTorch autograd graph from output tensors.

    Args:
        filename (Union[str, Path]): filename to save the graph to
        outputs (Union[Tensor, Sequence[Tensor]]): output tensor(s) of forward pass
        named_params (Union[Dict[str, Any], Iterator[Tuple[str, Parameter]]]): dict of params to label nodes with
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.

    Returns:
        str: The (possibly relative) path of the rendered file.
    """

    return get_autograd_dot_from_outputs(
        outputs=outputs,
        named_params=named_params,
        additional_params=additional_params,
        show_attrs=show_attrs,
        show_saved=show_saved,
        max_attr_chars=max_attr_chars,
    ).render(filename, format='svg', cleanup=True)


def save_autograd_graph_from_module(
        filename: Union[str, Path],
        module: nn.Module,
        *args: Tensor,
        additional_params: Optional[dict] = None,
        show_attrs: bool = True,
        show_saved: bool = True,
        max_attr_chars: int = 50,
        from_forward: bool = False,
        **kwargs: Tensor
) -> str:
    """Save Graphviz representation of PyTorch autograd graph from forward pass.

    Args:
        filename (Union[str, Path]): filename to save the graph to
        module (nn.Module): PyTorch model
        *args (Tensor): input to the model
        additional_params (dict): dict of additional params to label nodes with
        show_attrs (bool): whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved (bool): whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars (int): if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
        from_forward (bool): if True then use autograd graph otherwise analogvvn graph
        **kwargs (Tensor): input to the model

    Returns:
        str: The (possibly relative) path of the rendered file.
    """

    return get_autograd_dot_from_module(
        module,
        *args,
        additional_params=additional_params,
        show_attrs=show_attrs,
        show_saved=show_saved,
        max_attr_chars=max_attr_chars,
        from_forward=from_forward,
        **kwargs
    ).render(filename, format='svg', cleanup=True)


def save_autograd_graph_from_trace(filename: Union[str, Path], trace) -> str:
    """Save Graphviz representation of PyTorch autograd graph from trace.

    Args:
        filename (Union[str, Path]): filename to save the graph to
        trace (torch.jit.trace): the trace object to visualize.

    Returns:
        str: The (possibly relative) path of the rendered file.
    """

    return get_autograd_dot_from_trace(trace).render(filename, format='svg', cleanup=True)



from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Optional, Sequence, Tuple, Dict

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from analogvnn.nn.module.Layer import Layer
from analogvnn.nn.module.Model import Model

__all__ = ['TensorboardModelLog']

from analogvnn.utils.get_model_summaries import get_model_summaries


class TensorboardModelLog:
    """Tensorboard model log.

    Attributes:
        model (nn.Module): the model to log.
        tensorboard (SummaryWriter): the tensorboard.
        layer_data (bool): whether to log the layer data.
        _log_record (Dict[str, bool]): the log record.
    """

    model: nn.Module
    tensorboard: Optional[SummaryWriter]
    layer_data: bool
    _log_record: Dict[str, bool]

    def __init__(self, model: Model, log_dir: str):
        """Log the model to Tensorboard.

        Args:
            model (nn.Module): the model to log.
            log_dir (str): the directory to log to.
        """

        super().__init__()
        self.model = model
        self.tensorboard = None
        self.layer_data = True
        self._log_record = {}

        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        self.set_log_dir(log_dir)
        if hasattr(model, 'subscribe_tensorboard'):
            model.subscribe_tensorboard(tensorboard=self)

    def set_log_dir(self, log_dir: str) -> TensorboardModelLog:
        """Set the log directory.

        Args:
            log_dir (str): the log directory.

        Returns:
            TensorboardModelLog: self.

        Raises:
            ValueError: if the log directory is invalid.
        """

        # https://github.com/tensorflow/tensorboard/pull/6135
        from tensorboard.compat import tf
        if getattr(tf, 'io', None) is None:
            import tensorboard.compat.tensorflow_stub as new_tf
            tf.__dict__.update(new_tf.__dict__)

        if os.path.isdir(log_dir):
            self.tensorboard = SummaryWriter(log_dir=log_dir)
        else:
            raise ValueError(f'Log directory {log_dir} does not exist.')
        return self

    def _add_layer_data(self, epoch: int = None):
        """Add the layer data to the tensorboard.

        Args:
            epoch (int): the epoch to add the data for.
        """

        for name, parameter in self.model.named_parameters():
            if not parameter.requires_grad:
                continue

            self.tensorboard.add_histogram(name, parameter.data, epoch)

    def on_compile(self, layer_data: bool = True):
        """Called when the model is compiled.

        Args:
            layer_data (bool): whether to log the layer data.
        """

        if self.layer_data:
            self.layer_data = layer_data

        if self.layer_data:
            self._add_layer_data(epoch=-1)
        return self

    def add_graph(
            self,
            train_loader: DataLoader,
            model: Optional[nn.Module] = None,
            input_size: Optional[Sequence[int]] = None,
    ) -> TensorboardModelLog:
        """Add the model graph to the tensorboard.

        Args:
            train_loader (DataLoader): the train loader.
            model (Optional[nn.Module]): the model to log.
            input_size (Optional[Sequence[int]]): the input size.

        Returns:
            TensorboardModelLog: self.
        """

        if model is None:
            model = self.model

        log_id = f'{self.tensorboard.log_dir}_{TensorboardModelLog.add_graph.__name__}_{id(model)}'
        if log_id in self._log_record:
            return self

        if input_size is None:
            data_shape = next(iter(train_loader))[0].shape
            input_size = [1] + list(data_shape)[1:]

        use_autograd_graph = False
        if isinstance(model, Layer):
            use_autograd_graph = model.use_autograd_graph
            model.use_autograd_graph = False

        graph_path = Path(self.tensorboard.log_dir).joinpath(f'graph_{model.__class__.__name__}_{id(model)}')
        with SummaryWriter(log_dir=str(graph_path)) as graph_writer:
            graph_writer.add_graph(model, torch.zeros(input_size).to(model.device))

        self._log_record[log_id] = True
        if isinstance(model, Layer):
            model.use_autograd_graph = use_autograd_graph

        return self

    def add_summary(
            self,
            input_size: Optional[Sequence[int]] = None,
            train_loader: Optional[DataLoader] = None,
            model: Optional[nn.Module] = None,
            *args,
            **kwargs
    ) -> Tuple[str, str]:
        """Add the model summary to the tensorboard.

        Args:
            input_size (Optional[Sequence[int]]): the input size.
            train_loader (Optional[DataLoader]): the train loader.
            model (nn.Module): the model to log.
            *args: the arguments to torchinfo.summary.
            **kwargs: the keyword arguments to torchinfo.summary.

        Returns:
            Tuple[str, str]: the model __repr__ and the model summary.
        """

        if model is None:
            model = self.model

        log_id = f'{self.tensorboard.log_dir}_{TensorboardModelLog.add_summary.__name__}_{id(model)}'

        model_str, nn_model_summary = get_model_summaries(
            model=model,
            input_size=input_size,
            train_loader=train_loader,
            *args,  # noqa: B026
            **kwargs
        )

        if log_id in self._log_record:
            return model_str, nn_model_summary

        self.tensorboard.add_text(
            f'str ({model.__class__.__name__})',
            re.sub('\n', '\n    ', f'    {model_str}')
        )
        self.tensorboard.add_text(
            f'summary ({model.__class__.__name__})',
            re.sub('\n', '\n    ', f'    {nn_model_summary}')
        )
        self._log_record[log_id] = True
        return model_str, nn_model_summary

    def register_training(self, epoch: int, train_loss: float, train_accuracy: float) -> TensorboardModelLog:
        """Register the training data.

        Args:
            epoch (int): the epoch.
            train_loss (float): the training loss.
            train_accuracy (float): the training accuracy.

        Returns:
            TensorboardModelLog: self.
        """

        self.tensorboard.add_scalar('Loss/train', train_loss, epoch)
        self.tensorboard.add_scalar('Accuracy/train', train_accuracy, epoch)
        if self.layer_data:
            self._add_layer_data(epoch=epoch)
        return self

    def register_testing(self, epoch: int, test_loss: float, test_accuracy: float) -> TensorboardModelLog:
        """Register the testing data.

        Args:
            epoch (int): the epoch.
            test_loss (float): the test loss.
            test_accuracy (float): the test accuracy.

        Returns:
            TensorboardModelLog: self.
        """

        self.tensorboard.add_scalar('Loss/test', test_loss, epoch)
        self.tensorboard.add_scalar('Accuracy/test', test_accuracy, epoch)
        return self

    # noinspection PyUnusedLocal
    def close(self, *args, **kwargs):
        """Close the tensorboard.

        Args:
            *args: ignored.
            **kwargs: ignored.
        """

        if self.tensorboard is not None:
            self.tensorboard.close()
            self.tensorboard = None

    def __enter__(self):
        """Enter the TensorboardModelLog context.

        Returns:
            TensorboardModelLog: self.
        """

        return self

    __exit__ = close
    """Close the tensorboard."""



import numpy as np

from analogvnn.utils.common_types import TENSOR_OPERABLE

__all__ = ['gaussian_dirac_delta']


def gaussian_dirac_delta(x: TENSOR_OPERABLE, std: TENSOR_OPERABLE = 0.001) -> TENSOR_OPERABLE:
    """Gaussian Dirac Delta function with standard deviation `std`.

    Args:
        x (TENSOR_OPERABLE): Tensor
        std (TENSOR_OPERABLE): standard deviation.

    Returns:
        TENSOR_OPERABLE: TENSOR_OPERABLE with the same shape as x, with values of the Gaussian Dirac Delta function.
    """

    return 1 / (np.abs(std) * np.sqrt(np.pi)) * np.exp(-((x / std) ** 2))



from __future__ import annotations

from typing import Callable, Any

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn import ModuleDict
from torch.nn.utils import parametrize

__all__ = ['PseudoParameter']


class PseudoParameter(nn.Module):
    """A parameterized parameter which acts like a normal parameter during gradient updates.

    PyTorch's ParameterizedParameters vs AnalogVNN's PseudoParameters:

    - Similarity (Forward or Parameterizing the data):
        > Data -> ParameterizingModel -> Parameterized Data

    - Difference (Backward or Gradient Calculations):
      - ParameterizedParameters
        > Parameterized Data -> ParameterizingModel -> Data
      - PseudoParameters
        > Parameterized Data -> Data

    Attributes:
        _transformation (Callable): the transformation.
        _transformed (nn.Parameter): the transformed parameter.

    Properties:
        grad (Tensor): the gradient of the parameter.
        module (PseudoParameterModule): the module that wraps the parameter and the transformation.
        transformation (Callable): the transformation.
    """

    _transformation: Callable
    _transformed: nn.Parameter

    @staticmethod
    def identity(x: Any) -> Any:
        """The identity function.

        Args:
            x (Any): the input tensor.

        Returns:
            Any: the input tensor.
        """

        return x

    def __init__(self, data=None, requires_grad=True, transformation=None):
        """Initializes the parameter.

        Args:
            data: the data for the parameter.
            requires_grad (bool): whether the parameter requires gradient.
            transformation (Callable): the transformation.
        """

        super().__init__()
        self.original = nn.Parameter(data=data, requires_grad=requires_grad)
        self._transformed = nn.Parameter(data=data, requires_grad=requires_grad)
        self._transformed.original = self
        self._transformation = self.identity
        self.set_transformation(transformation)
        self.substitute_member(self.original, self._transformed, 'grad')

    def __call__(self, *args, **kwargs):
        """Transforms the parameter.

        Args:
            *args: additional arguments.
            **kwargs: additional keyword arguments.

        Returns:
            nn.Parameter: the transformed parameter.

        Raises:
            RuntimeError: if the transformation callable fails.
        """

        try:
            self._transformed.data = self._transformation(self.original)
        except Exception as e:
            raise RuntimeError(f'here: {e.args}') from e
        return self._transformed

    def set_original_data(self, data: Tensor) -> PseudoParameter:
        """Set data to the original parameter.

        Args:
            data (Tensor): the data to set.

        Returns:
            PseudoParameter: self.
        """

        self.original.data = data
        return self

    forward = __call__
    """Alias for __call__"""

    _call_impl = __call__
    """Alias for __call__"""

    right_inverse = set_original_data
    """Alias for set_original_data."""

    def __repr__(self):
        """Returns a string representation of the parameter.

        Returns:
            str: the string representation.
        """

        return f'{self.__class__.__name__}(' \
               f'transform={self.transformation}' \
               f', original={self.original}' \
               f')'

    @property
    def transformation(self):
        """Returns the transformation.

        Returns:
            Callable: the transformation.
        """

        return self._transformation

    @transformation.setter
    def transformation(self, transformation: Callable):
        """Sets the transformation.

        Args:
            transformation (Callable): the transformation.
        """

        self.set_transformation(transformation)

    def set_transformation(self, transformation) -> PseudoParameter:
        """Sets the transformation.

        Args:
            transformation (Callable): the transformation.

        Returns:
            PseudoParameter: self.
        """

        self._transformation = transformation
        if isinstance(self._transformation, nn.Module):
            self._transformation.eval()
        return self

    @staticmethod
    def substitute_member(
            tensor_from: Any,
            tensor_to: Any,
            property_name: str,
            setter: bool = True
    ):
        """Substitutes a member of a tensor as property of another tensor.

        Args:
            tensor_from (Any): the tensor property to substitute.
            tensor_to (Any): the tensor property to substitute to.
            property_name (str): the name of the property.
            setter (bool): whether to substitute the setter.
        """

        def getter_fn(self):
            return getattr(tensor_to, property_name)

        def setter_fn(self, value):
            setattr(tensor_to, property_name, value)

        new_class = type(tensor_from.__class__.__name__, (tensor_from.__class__,), {})

        if not setter:
            setattr(new_class, property_name, property(getter_fn))
        else:
            setattr(new_class, property_name, property(getter_fn, setter_fn))

        tensor_from.__class__ = new_class

    @classmethod
    def parameterize(cls, module: nn.Module, param_name: str, transformation: Callable) -> PseudoParameter:
        """Parameterizes a parameter.

        Args:
            module (nn.Module): the module.
            param_name (str): the name of the parameter.
            transformation (Callable): the transformation to apply.

        Returns:
            PseudoParameter: the parameterized parameter.
        """

        assert hasattr(module, param_name)

        param = getattr(module, param_name)
        new_param = cls(data=param, requires_grad=param.requires_grad, transformation=transformation)

        delattr(module, param_name)

        if not parametrize.is_parametrized(module):
            # Change the class
            parametrize._inject_new_class(module)
            # Inject a ``ModuleDict`` into the instance under module.parametrizations
            module.parametrizations = ModuleDict()

        module.parametrizations[param_name] = new_param
        parametrize._inject_property(module, param_name)
        return new_param

    @classmethod
    def parametrize_module(cls, module: nn.Module, transformation: Callable, requires_grad: bool = True):
        """Parametrize all parameters of a module.

        Args:
            module (nn.Module): the module parameters to parametrize.
            transformation (Callable): the transformation.
            requires_grad (bool): if True, only parametrized parameters that require gradients.
        """

        with torch.no_grad():
            for name, parameter in list(module.named_parameters(recurse=False)):
                if isinstance(parameter, cls):
                    continue

                if requires_grad and not parameter.requires_grad:
                    continue

                cls.parameterize(module=module, param_name=name, transformation=transformation)

            for sub_module in module.children():
                if sub_module == module:
                    continue
                if hasattr(module, 'parametrizations') and (
                        sub_module is module.parametrizations or sub_module in module.parametrizations
                ):
                    continue
                cls.parametrize_module(module=sub_module, transformation=transformation, requires_grad=requires_grad)



from __future__ import annotations

from typing import TypeVar, Optional

import torch
from torch import nn

from analogvnn.nn.module.Model import Model

T = TypeVar('T', bound=nn.Module)

__all__ = ['Sequential']


class Sequential(Model, nn.Sequential):
    """Base class for all sequential models."""

    def __call__(self, *args, **kwargs):
        """Call the model.

        Args:
            *args: The input.
            **kwargs: The input.

        Returns:
            torch.Tensor: The output of the model.
        """

        if not self._compiled:
            self.compile()

        return super().__call__(*args, **kwargs)

    def compile(self, device: Optional[torch.device] = None, layer_data: bool = True):
        """Compile the model and add forward graph.

        Args:
            device (torch.device): The device to run the model on.
            layer_data (bool): True if the data of the layers should be compiled.

        Returns:
            Sequential: self
        """
        arr = [self.graphs.INPUT, *list(self.registered_children()), self.graphs.OUTPUT]
        self.graphs.forward_graph.add_connection(*arr)
        return super().compile(device, layer_data)

    def add_sequence(self, *args):
        """Add a sequence of modules to the forward graph of model.

        Args:
            *args (nn.Module): The modules to add.
        """

        return self.extend(args)



from typing import Optional, Sequence, Tuple

from torch import nn
from torch.utils.data import DataLoader

from analogvnn.nn.module.Layer import Layer


def get_model_summaries(  # noqa: C901
        model: Optional[nn.Module],
        input_size: Optional[Sequence[int]] = None,
        train_loader: DataLoader = None,
        *args,
        **kwargs
) -> Tuple[str, str]:
    """Creates the model summaries.

    Args:
        train_loader (DataLoader): the train loader.
        model (nn.Module): the model to log.
        input_size (Optional[Sequence[int]]): the input size.
        *args: the arguments to torchinfo.summary.
        **kwargs: the keyword arguments to torchinfo.summary.

    Returns:
        Tuple[str, str]: the model __repr__ and the model summary.

    Raises:
        ImportError: if torchinfo (https://github.com/tyleryep/torchinfo) is not installed.
        ValueError: if the input_size and train_loader are None.
    """

    try:
        import torchinfo
    except ImportError as e:
        raise ImportError('requires torchinfo: https://github.com/tyleryep/torchinfo') from e

    if input_size is None and train_loader is None and 'input_size' not in kwargs:
        raise ValueError('input_size or train_loader must be provided')

    if 'input_size' not in kwargs:
        if input_size is None:
            data_shape = list(next(iter(train_loader))[0].shape)
            if train_loader.batch_size > 0:
                data_shape[0] = 1

            input_size = data_shape

        kwargs['input_size'] = input_size

    use_autograd_graph = False
    if isinstance(model, Layer):
        use_autograd_graph = model.use_autograd_graph
        model.use_autograd_graph = True

    if 'depth' not in kwargs:
        kwargs['depth'] = 10
    if 'col_names' not in kwargs:
        kwargs['col_names'] = tuple(e.value for e in torchinfo.ColumnSettings)
    if 'verbose' not in kwargs:
        kwargs['verbose'] = torchinfo.Verbosity.QUIET

    model_summary = torchinfo.summary(
        model,
        *args,
        **kwargs,
    )

    if isinstance(model, Layer):
        model.use_autograd_graph = use_autograd_graph

    model_summary.formatting.verbose = torchinfo.Verbosity.VERBOSE
    model_str = str(model)
    model_summary = f'{model_summary}'
    return model_str, model_summary



from __future__ import annotations

import typing

import networkx

from analogvnn.graph.GraphEnum import GraphEnum

if typing.TYPE_CHECKING:
    import graphviz

__all__ = ['to_graphviz_digraph']


def to_graphviz_digraph(from_graph: networkx.DiGraph, real_label: bool = False) -> graphviz.Digraph:
    """Returns a pygraphviz graph from a NetworkX graph N.

    Args:
        from_graph (networkx.DiGraph): the graph to convert.
        real_label (bool): True to use the real label.

    Returns:
        graphviz.Digraph: the converted graph.

    Raises:
        ImportError: if graphviz (https://pygraphviz.github.io/) is not available.
    """

    try:
        from graphviz import Digraph
    except ImportError as e:
        raise ImportError('requires graphviz: https://pygraphviz.github.io/') from e

    strict = networkx.number_of_selfloops(from_graph) == 0 and not from_graph.is_multigraph()
    node_attr = {
        'style': 'filled',
        'shape': 'box',
        'align': 'left',
        'fontsize': '10',
        'ranksep': '0.1',
        'height': '0.2',
        'fontname': 'monospace'
    }
    new_graph = Digraph(
        name=from_graph.name,
        strict=strict,
        node_attr=node_attr,
        format='svg'
    )

    # default graph attributes
    new_graph.graph_attr.update(from_graph.graph.get('graph', {}))
    new_graph.node_attr.update(from_graph.graph.get('node', {}))
    new_graph.edge_attr.update(from_graph.graph.get('edge', {}))

    new_graph.graph_attr.update(
        (k, v) for k, v in from_graph.graph.items() if k not in ('graph', 'node', 'edge')
    )

    # add nodes
    for n, nodedata in from_graph.nodes(data=True):
        attr = {k: str(v) for k, v in nodedata.items()}
        if isinstance(n, GraphEnum):
            attr['fillcolor'] = 'lightblue'
        attr['name'] = str(id(n))
        attr['label'] = str(n).strip()
        new_graph.node(**attr)

    for u, v, edgedata in from_graph.edges(data=True):
        attr = {k: str(v) for k, v in edgedata.items()}
        attr['tail_name'] = str(id(u))
        attr['head_name'] = str(id(v))

        if real_label and 'real_label' in attr:
            attr['label'] = attr['real_label']
        else:
            attr['label'] = attr['label'].replace('->', '→') if 'label' in attr else None
        new_graph.edge(**attr)

    return new_graph
